; 

FILEMAP STRUCT
    hFile           DWORD   ?
    hMap            DWORD   ?
    dwFileSizeLo    DWORD   ?
    dwFileSizeHi    DWORD   ?
    dwPageSize      DWORD   ?
    dwAllocAlign    DWORD   ?
    pView           DWORD   ?
FILEMAP ENDS

FILEMAPVIEW STRUCT
    pStart          DWORD   ?
    dwLength        DWORD   ?
    pMap            DWORD   ?
    pRealStart      DWORD   ?
    oRealStartLo    DWORD   ?
    oRealStartHi    DWORD   ?
    dwRealLength    DWORD   ?
    pNextView       DWORD   ?
    pLastView       DWORD   ?
FILEMAPVIEW ENDS

; 

FileMapOpen PROC USES esi spFilename:DWORD
IFE ASMRT_MAKE_INC
    LOCAL   sys:SYSTEM_INFO
    
    MemAlloc esi, SIZEOF FILEMAP
    
    invoke  CreateFile, spFilename, GENERIC_READ or GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0
    mov     [esi].FILEMAP.hFile, eax
    
    .if(eax == INVALID_HANDLE_VALUE)
        invoke  GetLastError
        cmp     eax, ERROR_FILE_NOT_FOUND
        jne     Failed_CreateFile
        
        invoke  CreateFile, spFilename, GENERIC_READ or GENERIC_WRITE, 0, 0, CREATE_NEW, 0, 0
        cmp     eax, INVALID_HANDLE_VALUE
        je      Failed_CreateFile
        
        mov     [esi].FILEMAP.hFile, eax
    .endif
    
    mov     [esi].FILEMAP.pView, 0
    
    invoke  GetSystemInfo, ADDR sys
    
    move    [esi].FILEMAP.dwAllocAlign, eax, sys.dwAllocationGranularity
    move    [esi].FILEMAP.dwPageSize,   eax, sys.dwPageSize
    
    invoke  GetFileSize, [esi].FILEMAP.hFile, ADDR [esi].FILEMAP.dwFileSizeHi
    .if([esi].FILEMAP.dwFileSizeHi == 0 && eax == 0)
        mov eax, sys.dwPageSize
    .endif
    mov     [esi].FILEMAP.dwFileSizeLo, eax
    
    invoke  CreateFileMapping, [esi].FILEMAP.hFile, 0, PAGE_READWRITE, [esi].FILEMAP.dwFileSizeHi, [esi].FILEMAP.dwFileSizeLo, 0
    or      eax, eax
    jz      Failed_CreateFileMapping
    mov     [esi].FILEMAP.hMap, eax
    
    mov     eax, esi
    ret

; ====================================
;   FAILED CODE
; ====================================

Failed_CreateFileMapping:
    invoke  CloseHandle, [esi].FILEMAP.hFile
Failed_CreateFile:
    MemFree esi
    xor     eax, eax
    ErrorThrow ASMRT_ERROR_CANNOT_OPEN_FILE
ENDIF
    ret
FileMapOpen ENDP

; -------------------------------------------------------------------------

FileMapClose PROC USES edi pMap:DWORD
IFE ASMRT_MAKE_INC
    mov     edi, pMap
    mov     edi, [edi].FILEMAP.pView
    .while(edi)
        .if([edi].FILEMAPVIEW.pStart)
            invoke  UnmapViewOfFile, [edi].FILEMAPVIEW.pRealStart
        .endif
        
        push    [edi].FILEMAPVIEW.pNextView
        MemFree edi
        pop     edi
    .endw
    
    mov     edi, pMap
    invoke  CloseHandle, [edi].FILEMAP.hMap
    invoke  CloseHandle, [edi].FILEMAP.hFile
    MemFree edi
    xor     eax, eax
ENDIF
    ret
FileMapClose ENDP

; 

FileMapLock PROC USES esi edi pMap:DWORD, dwOffsetHi:DWORD, dwOffsetLo:DWORD, dwSize:DWORD
IFE ASMRT_MAKE_INC
    LOCAL   oStart:DWORD
    
    MemAlloc esi, SIZEOF FILEMAPVIEW
    mov     edi, pMap
    mov     [esi].FILEMAPVIEW.pMap, edi
    
    mov     eax, dwOffsetLo
    mov     edx, dwOffsetHi
    mov     ecx, [edi].FILEMAP.dwAllocAlign
    div     ecx
    mov     eax, dwOffsetLo
    sub     eax, edx
    mov     oStart, edx
    mov     edx, dwOffsetHi
    mov     [esi].FILEMAPVIEW.oRealStartLo, eax
    mov     [esi].FILEMAPVIEW.oRealStartHi, edx
    
    mov     eax, dwSize
    mov     [esi].FILEMAPVIEW.dwLength, eax
    xor     edx, edx
    mov     ecx, [edi].FILEMAP.dwPageSize
    div     ecx
    add     eax, ecx
    mov     [esi].FILEMAPVIEW.dwRealLength, eax
    
    xor     edx, edx
    add     eax, dwOffsetLo
    adc     edx, dwOffsetHi
    sub     eax, oStart
    sbb     edx, 0
    
    .if((edx > [edi].FILEMAP.dwFileSizeHi) || (edx == [edi].FILEMAP.dwFileSizeHi && eax > [edi].FILEMAP.dwFileSizeLo))
        push    eax
        push    edx
        
        mov edx, [edi].FILEMAP.pView
        .while(edx)
            mov     eax, [edx].FILEMAPVIEW.pStart
            .if(eax)
                sub     eax, [edx].FILEMAPVIEW.pRealStart
                mov     [edx].FILEMAPVIEW.pStart, eax
                
                push    edx
                invoke  UnmapViewOfFile, [edx].FILEMAPVIEW.pRealStart
                pop     edx
            .endif
            
            mov     edx, [edx].FILEMAPVIEW.pNextView
        .endw
        
        invoke  CloseHandle, [edi].FILEMAP.hMap
        pop     edx
        pop     eax
        invoke  CreateFileMapping, [edi].FILEMAP.hFile, 0, PAGE_READWRITE, edx, eax, 0
        .if(!eax)
            invoke  CloseHandle, [edi].FILEMAP.hFile
            MemFree esi
            mov     esi, [edi].FILEMAP.pView
            .while(esi)
                push    [esi].FILEMAPVIEW.pNextView
                MemFree esi
                pop     esi
            .endw
            MemFree edi
            ErrorThrow ASMRT_ERROR_CANNOT_RESIZE_FILE
        .endif
        mov     [edi].FILEMAP.hMap, eax
        mov     edx, [esi].FILEMAP.pView
        .while(edx)
            push    edx
            invoke  MapViewOfFile, [edi].FILEMAP.hMap, FILE_MAP_WRITE, [edx].FILEMAPVIEW.oRealStartHi, [edx].FILEMAPVIEW.oRealStartLo, [edx].FILEMAPVIEW.dwRealLength
            pop     edx
            .if(eax)
                mov     [edx].FILEMAPVIEW.pRealStart, eax
                add     [edx].FILEMAPVIEW.pStart, eax
            .else
                mov     [edx].FILEMAPVIEW.pStart, 0
            .endif
            
            mov     edx, [edx].FILEMAPVIEW.pNextView
        .endw
    .endif
    
    invoke  MapViewOfFile, [edi].FILEMAP.hMap, FILE_MAP_WRITE, [esi].FILEMAPVIEW.oRealStartHi, [esi].FILEMAPVIEW.oRealStartLo, [esi].FILEMAPVIEW.dwRealLength
    .if(eax)
        mov     [esi].FILEMAPVIEW.pRealStart, eax
        add     eax, oStart
        mov     [esi].FILEMAPVIEW.pStart, eax
        
        mov     eax, [edi].FILEMAP.pView
        .if(eax)
            .while([eax].FILEMAPVIEW.pNextView)
                mov eax, [eax].FILEMAPVIEW.pNextView
            .endw
            mov     [eax].FILEMAPVIEW.pNextView, esi
        .else
            mov     [edi].FILEMAP.pView, esi
        .endif
        
        mov     [esi].FILEMAPVIEW.pLastView, eax
        mov     [esi].FILEMAPVIEW.pNextView, 0
        mov     eax, esi
    .else
        MemFree esi
        xor     eax, eax
        ErrorThrow ASMRT_ERROR_CANNOT_OPEN_FILE
    .endif
ENDIF
    ret
FileMapLock ENDP

; -------------------------------------------------------------------------

FileMapUnlock PROC pView:DWORD
IFE ASMRT_MAKE_INC
    mov     eax, pView
    mov     ecx, [eax].FILEMAPVIEW.pLastView
    mov     edx, [eax].FILEMAPVIEW.pNextView
    
    .if(ecx)
        mov     [ecx].FILEMAPVIEW.pNextView, edx
    .else
        mov     ecx, [eax].FILEMAPVIEW.pMap
        mov     [ecx].FILEMAP.pView, edx
    .endif
    .if(edx)
        mov     [edx].FILEMAPVIEW.pLastView, ecx
    .endif
    
    .if([eax].FILEMAPVIEW.pStart)
        invoke  UnmapViewOfFile, [eax].FILEMAPVIEW.pRealStart
    .endif
    
    MemFree pView
    
    xor     eax, eax
ENDIF
    ret
FileMapUnlock ENDP

; 
